//
//  OACalloutView.m
//  ObjCAdditions
//
// Copyright 2011 A25 SIA
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
//
// Based on https://github.com/edanuff/MonoTouchCalloutView


#import "OACalloutView.h"
#import <QuartzCore/QuartzCore.h>

#define CENTER_IMAGE_WIDTH 31
#define CALLOUT_HEIGHT 70
#define MIN_LEFT_IMAGE_WIDTH 16
#define MIN_RIGHT_IMAGE_WIDTH 16
#define ANCHOR_X 32
#define ANCHOR_Y 60
#define CENTER_IMAGE_ANCHOR_OFFSET_X 15
#define MIN_ANCHOR_X MIN_LEFT_IMAGE_WIDTH + CENTER_IMAGE_ANCHOR_OFFSET_X
#define TITLE_FONT_SIZE 16.0f
#define SUBTITLE_FONT_SIZE 12.0f


@interface OACalloutView ()

@property (nonatomic,retain) UIImage *calloutLeftImage;
@property (nonatomic,retain) UIImage *calloutCenterImage;
@property (nonatomic,retain) UIImage *calloutRightImage;
@property (nonatomic,retain) UIImageView *calloutLeft;
@property (nonatomic,retain) UIImageView *calloutCenter;
@property (nonatomic,retain) UIImageView *calloutRight;
@property (nonatomic,retain) UILabel *calloutTitleLabel;
@property (nonatomic,retain) UILabel *calloutSubtitleLabel;
@property (nonatomic,retain) UIButton *calloutButton;

- (void)prepareView;
- (void)updateLayout:(BOOL)animated;
- (void)showAnimationWillStart:(NSString *)animationID context:(void *)context;
- (void)showAnimationDidStop:(NSString *)animationID finished:(NSNumber *)finished context:(void *)context;

@end


@implementation OACalloutView

#pragma mark -
#pragma mark Synthesized properties

@synthesize anchorPoint;
@synthesize calloutButton;
@synthesize calloutCenter;
@synthesize calloutCenterImage;
@synthesize calloutLeft;
@synthesize calloutLeftImage;
@synthesize calloutRight;
@synthesize calloutRightImage;
@synthesize calloutSubtitleLabel;
@synthesize calloutTitleLabel;
@synthesize labelHeight;
@synthesize labelOriginY;
@synthesize maxWidth;

#pragma mark -
#pragma mark Memory managment

- (id)initWithLeftImage:(UIImage *)leftImage centerImage:(UIImage *)centerImage rightImage:(UIImage *)rightImage {
	self = [super init];
	if (self) {
		self.calloutLeftImage = [leftImage stretchableImageWithLeftCapWidth:leftImage.size.width - 1 topCapHeight:0];
		self.calloutCenterImage = centerImage;
		self.calloutRightImage = [rightImage stretchableImageWithLeftCapWidth:1 topCapHeight:0];
		self.maxWidth = 320.f;
		[self prepareView];
	}
	return self;
}

- (void)prepareView {
	self.backgroundColor = [UIColor clearColor];
	self.opaque = NO;
	
	self.calloutLeft = [[UIImageView alloc] init];
	self.calloutLeft.image = self.calloutLeftImage;
	[self addSubview:self.calloutLeft];

	self.calloutCenter = [[UIImageView alloc] init];
	self.calloutCenter.image = self.calloutCenterImage;
	[self addSubview:self.calloutCenter];
		
	self.calloutRight = [[UIImageView alloc] init];
	self.calloutRight.image = self.calloutRightImage;
	[self addSubview:self.calloutRight];

	self.calloutTitleLabel = [[UILabel alloc] init];
	self.calloutTitleLabel.font = [UIFont boldSystemFontOfSize:TITLE_FONT_SIZE];
	self.calloutTitleLabel.textColor = [UIColor whiteColor];
	self.calloutTitleLabel.backgroundColor = [UIColor clearColor];
	
	self.calloutSubtitleLabel = [[UILabel alloc] init];
	self.calloutSubtitleLabel.font = [UIFont systemFontOfSize:SUBTITLE_FONT_SIZE];
	self.calloutSubtitleLabel.textColor = [UIColor whiteColor];
	self.calloutSubtitleLabel.backgroundColor = [UIColor clearColor];
	
	self.calloutButton = [UIButton buttonWithType:UIButtonTypeDetailDisclosure];
	self.calloutButton.adjustsImageWhenHighlighted = NO;
    self.rightView = self.calloutButton;
}


#pragma mark -
#pragma mark UIView lifecycle

- (void)updateLayout
{
	if (animationInProgress) {
		layoutUpdateNeeded = YES;
	} else {
		layoutUpdateNeeded = NO;
		[self updateLayout:YES];
	}
}

- (void)updateLayout:(BOOL)animated {
	animationInProgress = YES;

	if (animated && self.superview) {
		[UIView beginAnimations:nil context:nil];
		[UIView setAnimationBeginsFromCurrentState:YES];
		[UIView setAnimationDelegate:self];
		[UIView setAnimationWillStartSelector:@selector(showAnimationWillStart:context:)];
		[UIView setAnimationDidStopSelector:@selector(showAnimationDidStop:finished:context:)];
		[UIView setAnimationDuration:0.1f];
	}

	CGSize titleSize = [self.calloutTitleLabel.text length] ? [self.calloutTitleLabel.text sizeWithFont:self.calloutTitleLabel.font] : CGSizeZero;
	CGSize subtitleSize = [self.calloutSubtitleLabel.text length] ? [self.calloutSubtitleLabel.text sizeWithFont:self.calloutSubtitleLabel.font] : CGSizeZero;
	
	CGFloat maxLabelWidth = self.maxWidth - (MIN_LEFT_IMAGE_WIDTH + self.calloutButton.frame.size.width + MIN_RIGHT_IMAGE_WIDTH + 3.0f);
	CGFloat labelWidth = MAX(titleSize.width, subtitleSize.width);
	labelWidth = MIN(labelWidth, maxLabelWidth);
	CGFloat titleOffset = self.labelOriginY + floorf((self.labelHeight - titleSize.height - subtitleSize.height) / 2.0f);
				   
	CGRect frame = self.frame;
	frame.size.height = CALLOUT_HEIGHT;
	frame.size.width = labelWidth + MIN_LEFT_IMAGE_WIDTH + self.calloutButton.frame.size.width + MIN_RIGHT_IMAGE_WIDTH + 3.0f;
	self.frame = frame;	
	
	CGFloat leftWidth = ceilf((frame.size.width - calloutCenterImage.size.width) / 2.0f);
	CGFloat rightWidth = frame.size.width - leftWidth - calloutCenterImage.size.width;
	
	CGRect titleLabelFrame = CGRectMake(MIN_LEFT_IMAGE_WIDTH, titleOffset, labelWidth, titleSize.height);
	CGRect subtitleLabelFrame = CGRectMake(MIN_LEFT_IMAGE_WIDTH, titleOffset + titleSize.height, labelWidth, subtitleSize.height);
	
	self.calloutLeft.frame = CGRectMake(0.0f, 0.0f, leftWidth, CALLOUT_HEIGHT);
	self.calloutCenter.frame = CGRectMake(leftWidth, 0.0f, CENTER_IMAGE_WIDTH, CALLOUT_HEIGHT);
	self.calloutRight.frame = CGRectMake(leftWidth + CENTER_IMAGE_WIDTH, 0.0f, rightWidth, CALLOUT_HEIGHT);
	self.calloutTitleLabel.frame = titleLabelFrame;
	self.calloutSubtitleLabel.frame = subtitleLabelFrame;
	
	CGRect rightViewFrame = self.rightView.frame;
	rightViewFrame.origin.x = frame.size.width - rightViewFrame.size.width - MIN_RIGHT_IMAGE_WIDTH + 4.0f;
	rightViewFrame.origin.y = self.labelOriginY + floorf((self.labelHeight - rightViewFrame.size.height) / 2.0f);
	self.rightView.frame = rightViewFrame;
	
	self.layer.anchorPoint = CGPointMake(0.5f, ANCHOR_Y / self.frame.size.height);
	self.center = anchorPoint;
	
	if (animated && self.superview) {
		[UIView commitAnimations];
	} else {
		animationInProgress = NO;
	}
}

   
#pragma mark -
#pragma mark Control methods

- (void)showAnimated:(UIView *)parent {
	animationInProgress = YES;
	
	self.transform = CGAffineTransformMakeScale(0.8f, 0.8f);
	
	[UIView beginAnimations:nil context:nil];
	[UIView setAnimationDelegate:self];
	[UIView setAnimationWillStartSelector:@selector(showAnimationWillStart:context:)];
	[UIView setAnimationDidStopSelector:@selector(showAnimationDidStop:finished:context:)];
	[UIView setAnimationDuration:0.1f];
	[parent addSubview:self];
	
	self.transform = CGAffineTransformMakeScale(1.2f, 1.2f);
	
	[UIView commitAnimations];
}

#pragma mark -
#pragma mark Animation handlers

- (void)showAnimationWillStart:(NSString *)animationID context:(void *)context {
	animationInProgress = YES;
}

- (void)showAnimationDidStop:(NSString *)animationID finished:(NSNumber *)finished context:(void *)context {
	animationInProgress = NO;
	self.transform = CGAffineTransformIdentity;
	
	if (layoutUpdateNeeded) {
		[self updateLayout];
	}
}


#pragma mark -
#pragma mark Properties

- (void)setTitle:(NSString *)title {
	if ([title length]) {
		self.calloutTitleLabel.text = title;
		[self addSubview:self.calloutTitleLabel];
	} else {
		self.calloutTitleLabel.text = nil;
		[self.calloutTitleLabel removeFromSuperview];
	}
	[self updateLayout];
}

- (NSString *)title {
	return self.calloutTitleLabel.text;
}

- (void)setSubtitle:(NSString *)subtitle {
	if ([subtitle length]) {
		self.calloutSubtitleLabel.text = subtitle;
		[self addSubview:self.calloutSubtitleLabel];
	} else {
		self.calloutSubtitleLabel.text = nil;
		[self.calloutSubtitleLabel removeFromSuperview];
	}
	[self updateLayout];
}

- (NSString *)subtitle {
	return self.calloutSubtitleLabel.text;
}

- (void)setAnchorPoint:(CGPoint)point {
	anchorPoint = point;
	self.center = anchorPoint;
}

- (void)addButtonTarget:(id)target action:(SEL)action {
	[self.calloutButton addTarget:target action:action forControlEvents:UIControlEventTouchUpInside];
}

- (void)setRightView:(UIView *)rightView
{
    if (_rightView != rightView) {
        [_rightView removeFromSuperview];
        _rightView = rightView;
        
        [self addSubview:rightView];
        [self updateLayout:NO];
    }
}

- (BOOL)pointInside:(CGPoint)point withEvent:(UIEvent *)event {
	return self.rightView && CGRectContainsPoint(self.rightView.frame, point);
}

@end
